/*
 * Copyright (c) 2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/time.h>
#include <err.h>
#include <math.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "mor/socketcan.h"
#include "mor/harmonica.h"
#include "mor/relayboard.h"
#include "mor/morLib.h"

#include "mom.h"
#include "neoTime.h"
#include "mortimer.h"

static double wheelT(MOR_STR *, int, double, double, bool);
static void *morReadTask(void *);

/*
 * Init MOR platform
 *
 * Establish connection with the low-level hardware and set it to a known state
 *
 */

static
int
morInitBase(MOR_STR *mor)
{
	int i;

	mor->wheels = (MOR_WHEEL_CTRL_STR *)calloc(MOR_N_WHEELS,
	    sizeof(MOR_WHEEL_CTRL_STR));
	if (mor->wheels == NULL) {
		return -1;
	}
	for (i = 0; i < MOR_N_WHEELS; i++) {
		mor->wheels[i].steer = harmonicaInit(mor->canfd,
		    i*2+1, MOTIONTYPE_VELCTRL,
		    momWheels[i].steer.canOpen,
		    momWheels[i].steer.params);

		mor->wheels[i].drive = harmonicaInit(mor->canfd,
		    i*2+2, MOTIONTYPE_VELCTRL,
		    momWheels[i].drive.canOpen,
		    momWheels[i].drive.params);
		mor->wheels[i].geom = (MOR_SPEED_CTRL_GEOM_STR *)
		    malloc(sizeof(MOR_SPEED_CTRL_GEOM_STR));
		if (mor->wheels[i].steer == NULL
		    || mor->wheels[i].drive == NULL
		    || mor->wheels[i].geom == NULL) {
			fprintf(stderr, "%s: failed to initialize wheel %d\n",
			    __func__, i);
			return -1;
		}
	}
	return 0;
}

MOR_STR *
morInit(const char *canDev, const char *serialDev, bool initBase, bool initSpine)
{
	MOR_STR *mor;

	mor = (MOR_STR *)calloc(1, sizeof(MOR_STR));
	if (mor == NULL) {
		return NULL;
	}
	/* CAN Bus */
	mor->canfd = socketcanInit(canDev);
	if (mor->canfd == -1) {
		free(mor);
		return NULL;
	}
	/* Relayboard  */
	if (serialDev != NULL)
		mor->relayboard = relayboardInit(serialDev);
	else
		mor->relayboard = NULL;
	if (mor->relayboard == NULL) {
		fprintf(stderr, "%s: relayboard not available\n", __func__);
	}
	mor->odo = (MOR_ODO_STR *)calloc(1, sizeof(MOR_ODO_STR));
	if (mor->odo == NULL) {
		morEnd(mor);
		return NULL;
	}
	if (initBase) {
		if (morInitBase(mor) == -1) {
			morEnd(mor);
			return NULL;
		}
	}
	if (initSpine) {
		mor->spine  = harmonicaInit(mor->canfd, 0, MOTIONTYPE_POSCTRL,
		    &spineAddr, &spineParams);
		if (mor->spine == NULL) {
			morEnd(mor);
			return NULL;
		}
	}
	mor->initSpine = initSpine;
	mor->initBase = initBase;
	return mor;
}

/*
 * De-init MOR platform
 */
void
morEnd(MOR_STR *mor)
{
	int i;

	mor->dying = true;

	if (mor->readTask != 0)
		pthread_join(mor->readTask, NULL);

	if (mor->initBase) {
		for (i = 0; i < MOR_N_WHEELS; i++) {
			harmonicaEnd(mor->wheels[i].steer);
			harmonicaEnd(mor->wheels[i].drive);
			free(mor->wheels[i].geom);
		}
		free(mor->wheels);
	}
	if (mor->initSpine)
		harmonicaEnd(mor->spine);
	socketcanEnd(mor->canfd);
	free(mor->odo);
	free(mor);
}

/*
 * Perform actual initialisation of the MOR platform
 *
 * Start motor controllers - but don't move (yet)
 */
int
morInitPlatform(MOR_STR *mor)
{
	int i;

	/* init the watchdogs, but disable them */
	if (mor->initSpine) {
		harmonicaStartWatchdog(mor->spine, true);
		harmonicaStartWatchdog(mor->spine, false);
	}

	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaStartWatchdog(mor->wheels[i].steer, true);
		harmonicaStartWatchdog(mor->wheels[i].drive, true);
	}
	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaStartWatchdog(mor->wheels[i].steer, false);
		harmonicaStartWatchdog(mor->wheels[i].drive, false);
	}
	printf("Watchdogs initialized\n");

	if (mor->initSpine) {
		if (harmonicaInitCtrl(mor->spine) == -1) {
			printf("error initCtrl spine drive \n");
			return -1;
		}
		if (harmonicaStart(mor->spine) == -1) {
			printf("error starting spine drive\n");
			return -1;
		}
		harmonicaSetGearVelRadS(mor->spine, 0.0);
		printf("Spine motor started\n");
	}
	if (!mor->initBase)
		return 0;

	for (i = 0; i < MOR_N_WHEELS; i++) {

		if (harmonicaInitCtrl(mor->wheels[i].drive) == -1) {
			printf("error initCtrl drive %d\n",
			    mor->wheels[i].drive->driveParam.driveIdent);
			return -1;
		}
		if (harmonicaInitCtrl(mor->wheels[i].steer) == -1) {
			printf("error initCtrl steer %d\n",
			    mor->wheels[i].steer->driveParam.driveIdent);
			return -1;
		}
		mssleep(10);
		if (harmonicaStart(mor->wheels[i].drive) == -1) {
			printf("error start drive %d\n",
			    mor->wheels[i].drive->driveParam.driveIdent);
			return -1;
		}
		if (harmonicaStart(mor->wheels[i].steer) == -1) {
			printf("error start steer %d\n",
			    mor->wheels[i].drive->driveParam.driveIdent);
			return -1;
		}
		mssleep(10);
		harmonicaSetGearVelRadS(mor->wheels[i].steer, 0.0);
		harmonicaSetGearVelRadS(mor->wheels[i].drive, 0.0);
	}
	printf("Base motors started\n");
	return 0;
}

/*
 * Receive and evaluate messages on the CAN bus
 */
int
morEvalCanBuffer(MOR_STR *mor, int usec)
{
	CAN_MSG m;
	bool decoded = false;
	int i;

	if (socketcanReceiveMsgWait(mor->canfd, &m, usec) != 0)
		return -1;

	if (mor->initSpine)
		decoded = harmonicaDecode(mor->spine, &m);
	if (decoded)
		return 0;

	mor->motorFailure = false;
	for (i = 0; i < MOR_N_WHEELS; i++) {
		decoded |= harmonicaDecode(mor->wheels[i].steer, &m);
		decoded |= harmonicaDecode(mor->wheels[i].drive, &m);
		if (harmonicaGetMotorState(mor->wheels[i].steer)
		    == ST_MOTOR_FAILURE
		    || harmonicaGetMotorState(mor->wheels[i].drive)
		    == ST_MOTOR_FAILURE)
			mor->motorFailure = true;

		if (decoded)
			return 0;
	}
	printf("%s: unknown message from 0x%04x: 0x%02x, %02x\n", __func__,
	    m.id, m.data[0], m.data[1]);
	return 0;
}

/*
 * Perform homing of the platform (spine + 4 wheels)
 *
 * homing of spine is controlled by homeSpine
 */
int
morHoming(MOR_STR *mor)
{
	CAN_MSG m;
	int i, count, good;
	double homeVeloRadS = -1;
	double factorVel[MOR_N_WHEELS];

	if (mor->initSpine) {
		harmonicaInitHoming(mor->spine);
		if (harmonicaExecHoming(mor->spine) == -1) {
			return -1;
		}
		printf("Spine homing OK\n");
	}

	if (!mor->initBase)
		return 0;

	for (i = 0; i < MOR_N_WHEELS; i++) {
		factorVel[i] = - couplingFactor +
		    momWheel[i].arm/momWheel[i].radius;
		mor->wheels[i].geom->orient = momWheel[i].initOrient;
	}
	for (i = 0; i < MOR_N_WHEELS; i++)
		harmonicaInitHoming(mor->wheels[i].steer);
	/* start moving wheels */
	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaSetGearVelRadS(mor->wheels[i].steer, homeVeloRadS);
		harmonicaSetGearVelRadS(mor->wheels[i].drive,
		    homeVeloRadS * factorVel[i]);
	}
	mssleep(500);
	/* flush CAN bus */
	while (socketcanReceiveMsg(mor->canfd, &m) == 0)
		;

	/* Arm homing */
	for (i = 0; i < MOR_N_WHEELS; i++)
		intprtSetInt(mor->wheels[i].steer, 8, 'H', 'M', 1, 1);
	count = 0;
	do {
		for (i = 0; i < MOR_N_WHEELS; i++) {
			if (!mor->wheels[i].steer->homePos) {
				/* request homing status */
				intprtSetInt(mor->wheels[i].steer,
				    4, 'H', 'M', 1, 0);
				harmonicaSetGearVelRadS(mor->wheels[i].steer,
				    homeVeloRadS);
				harmonicaSetGearVelRadS(mor->wheels[i].drive,
				    homeVeloRadS * factorVel[i]);
			}
		}
		if (mor->initSpine)
			harmonicaSetGearVelRadS(mor->spine, 0.0);
		while (morEvalCanBuffer(mor, 1) == 0)
			;
		good = 0;
		for (i = 0; i < MOR_N_WHEELS; i++) {
			if (mor->wheels[i].steer->homePos) {
				good++;
				harmonicaSetGearPosVelRadS(mor->wheels[i].steer,
				    0.0, 0.0);
				harmonicaSetGearVelRadS(mor->wheels[i].drive,
				    0.0);
			}
		}
		if (good == MOR_N_WHEELS)
			break;
		mssleep(5);
		count++;

	} while (count < 20000);
	if (count == 20000)
		return -1;

	printf("Homing done\n");
	/* enable watchdogs */
	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaStartWatchdog(mor->wheels[i].steer, true);
		harmonicaStartWatchdog(mor->wheels[i].drive, true);
	}
	morStop(mor);
	return 0;
}


/*
 * angleLimit - limits angle between -M_PI and M_PI
 */
static double
morAngleLimit(double angle)
{
	if (angle > 0)
		return fmod(angle+M_PI, 2.0*M_PI) - M_PI;
	else
		return fmod(angle-M_PI, 2.0*M_PI) + M_PI;
}

/*
 * Generic platform motion control
 * Let 0,0 the robot center. Cx,Cy the wheel orientation axis.
 * Vx,Vy the robot velocity at 0,0. VCx, VCy the velocity at C.
 * W the robot rotation rate at 0.
 * R the wheel radius, S the wheel spin rate. T the wheel orientation and dT
 * the time derivative of T. ARMx,y the vector joining the wheel center to the
 * wheel orientation axis, 'arm' the length of this vector.
 *
 * We have:
 * VCx = Vx - Cy * W
 * VCy = Vy + Cx * W
 * And:
 * VCx = (R * S - arm.dT).cosT
 * VCy = (R * S - arm.dT).sinT
 *
 * Thus:
 * T = atan2(VCy, VCx) = atan2(Vy + Cx * W, Vx - Cy * W)
 * arm * dT = (R * S - VCx/cosT) = (R * S - VCy/sinT)
 *
 * Derivating T = atan2(..) gives the expression of dT in function of Vx,Vy,W
 * and their time derivatives, so we can limit dT to the max value by
 * interpolating a new Vx,Vy,W that do not violate the constraints.
 *
 * Then we have the new orientation T thanks to atan2, that gives a
 * dT within the limits by definition and finally S (wheel spin rate)
 * from the average of the two eq. with cosT or sinT to use the fact that
 * cosT+sinT not close from 0.
 */
void
morMove(MOR_STR *mor, double vx, double vy, double w, double servoperiod)
{
	MOR_SPEED_CTRL_GEOM_STR *wheel;
	double time;
	double n, m, nvx, nvy, nw, vcx, vcy, T, dT, S;
	double steerPos[MOR_N_WHEELS];
	static double pvx = 0.0, pvy = 0.0, pw = 0.0;
	int i, iter;

	/*
	 * compute time scaling factor to have dT in the limits for all wheels
	 * ('time' is in $servoperiod units)
	 */
	time = 1.0;
	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaGetPosRad(mor->wheels[i].steer, &steerPos[i]);
		steerPos[i] = morAngleLimit(steerPos[i]);

		wheel = mor->wheels[i].geom;
		/* Vc for previous Vx, Vy, W */
		vcx = pvx - momWheel[i].y*pw;
		vcy = pvy + momWheel[i].x*pw;
		n = vcx*vcx + vcy*vcy;

		if (n < 1e-6) {
			/*
			 * singularity when CIR is above one wheel (C)
			 * or the velocity was 0.
			 * We need to reach a new T before moving.
			 */
			T = wheelT(mor, i, vx - momWheel[i].y*w,
			    vy + momWheel[i].x*w, false);
			dT = T - steerPos[i];
			if (dT > M_PI)
				dT -=  2*M_PI;
			else if (dT < -M_PI)
				dT += 2*M_PI;
			dT /= servoperiod;

			if (dT > momWheel[i].dorientmax)
				dT = momWheel[i].dorientmax;
			else if (dT < -momWheel[i].dorientmax)
				dT = -momWheel[i].dorientmax;

			wheel->dorient = dT;
#if 0
			wheel->orient = steerPos[i] + dT*servoperiod;
			if (wheel->orient > M_PI)
				wheel->orient -= 2*M_PI;
			else if (wheel->orient < -M_PI)
				wheel->orient += 2*M_PI;
#else
			wheel->orient = T;
#endif
			/* don't change global velocity until
			   we reach requested T */
			if (fabs(dT)*servoperiod > 1e-2) {
				// printf("%d not moving dT: %.4lf\n",
				//  i, dT*servoperiod);
				time = 0;
			}
			continue;
		}
		/* we won't move while rotating wheels,
		   don't bother with the rest */
		if (time < 1e-5)
			continue;

		/* limit dT: iterate on the tangent of T = f(Vx,Vy,W) */
		for (iter = 5; iter > 0; iter--) {
			/* get new speed at time time */
			nvx = pvx + (vx - pvx)/time;
			nvy = pvy + (vy - pvy)/time;
			nw = pw + (w - pw)/time;

			/* Vc at mid-point between previous Vx,Vy,W
			   and new Vx,Vy,W */
			vcx = (pvx+nvx)/2. - momWheel[i].y*(pw+nw)/2.;
			vcy = (pvy+nvy)/2. + momWheel[i].x*(pw+nw)/2.;

			/* derivative of T = atan(Vcy/Vcx) = m/n
			   at mid-point */
			m = vcx*(nvy-pvy) - vcy*(nvx-pvx) +
			    (momWheel[i].x*(pvx+nvx)/2. +
				momWheel[i].y*(pvy+nvy)/2)*(nw-pw);
			n = vcx*vcx + vcy*vcy;

			/* check limit overshoot, tolerate 1e-3% overshoot */
			if (fabs(m) < (1+1e-5) * momWheel[i].dorientmax
			    * servoperiod * n)
				break;
			/* limit dT by increasing the time
			   to reach new Vx,Vy,W */
			time = time*fabs(m)/n/servoperiod
			    /momWheel[i].dorientmax;
		}
	}

	/* interpolate new speed according to the max factor for all wheels */
	if (time < 1e-5) {
		vx = pvx;
		vy = pvy;
		w = pw;
	} else if (time > 1.) {
		vx = pvx + (vx - pvx)/time;
		vy = pvy + (vy - pvy)/time;
		w = pw + (w - pw)/time;
	}

	/* get T and S according to new Vx,Vy,W */
	for (i = 0; i < MOR_N_WHEELS; i++) {
		wheel = mor->wheels[i].geom;
		/* Vc at the current Vx,Vy,W */
		vcx = vx - momWheel[i].y*w;
		vcy = vy + momWheel[i].x*w;
		/* T */
		if (vcx*vcx + vcy*vcy < 1e-6) {
			/* singularity when CIR is above one wheel (C)
			 * or the velocity was 0.
			 * T is undefined at 0,0,0
			 * but was already set above to the new target. */
			T = wheel->orient;
			dT = wheel->dorient;
		} else {
			T = wheelT(mor, i, vcx, vcy, true);
			dT = T - steerPos[i];
			if (dT > M_PI)
				dT -= 2*M_PI;
			else if (dT < -M_PI)
				dT += 2*M_PI;
			dT = dT/servoperiod;
			if (fabs(dT) > momWheel[i].dorientmax)
				fprintf(stderr, "wheel %d exceeds "
				    "dorientmax: %.3lf\n", i, dT);
		}

		/* S (take the best eq. where cosT != 0 or sinT != 0) */
		if (fabs(cos(T)) < fabs(sin(T)))
			S = vcy/sin(T) + momWheel[i].arm*w;
		else
			S = vcx/cos(T) + momWheel[i].arm*w;
		S = S / momWheel[i].radius;

		wheel->orient = T;
		wheel->dorient = dT;
		wheel->spin = S;
		// printf("  %d %.3lf %.3lf %.3lf\n", i, T, dT, S);
	}
	/* Remember last state */
	pvx = vx;
	pvy = vy;
	pw = w;
}

/*
 *  return wheel orientation
 */
static double
wheelT(MOR_STR *mor, int i, double vcx, double vcy, bool closest)
{
	MOR_SPEED_CTRL_GEOM_STR *wheel = mor->wheels[i].geom;
	double T;

	T = atan2(vcy, vcx);
	if (closest) {
		/*  when moving, must take the closest solution */
		if (fabs(morAngleLimit(T+M_PI-wheel->orient))
		    < fabs(morAngleLimit(T-wheel->orient)))
			T += M_PI;
		else if (fabs(morAngleLimit(T-M_PI-wheel->orient)) <
		    fabs(morAngleLimit(T-wheel->orient)))
			T -= M_PI;
	} else {
		/* when not moving, take the nicest solution */
		switch (i) {
		case 0:
			if (T > -3*M_PI_4 && T < M_PI_4)
				T += M_PI;
			break;
		case 1:
			if (T < 3*M_PI_4 && T > -M_PI_4)
				T += M_PI;
			break;
		case 2:
			if (T < -3*M_PI_4 || T > M_PI_4)
				T += M_PI;
			break;
		case 3:
			if (T < -M_PI_4 || T > 3*M_PI_4)
				T += M_PI;
		}
		if (T > M_PI)
			T -= 2*M_PI;
	}
	return T;
}

/*
 * Odometry
 *
 * We have
 * Vx - Y.W = VCx = (RS - arm.(dT + W)).cosT
 * Vy + X.W = VCy = (RS - arm.(dT + W)).sinT
 *
 * Summing VCx(0) - VCx(1) + VCx(2) - VCx(3) and VCy(0..3) will give W with a
 * k factor.
 */
void
morOdo(MOR_STR *mor, double servoperiod)
{
	double vt, k, vx, vy, dt;
	double driveDeltaPos, driveVel, steerPos[MOR_N_WHEELS], steerDeltaPos, steerVel;
	double rsmldt[MOR_N_WHEELS];
	int i;

	vt = 0.;
	k = 0.;
	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaGetDeltaPosVelRadS(mor->wheels[i].drive,
		    &driveDeltaPos, &driveVel);
		harmonicaGetDeltaPosVelRadS(mor->wheels[i].steer,
		    &steerDeltaPos, &steerVel);
		harmonicaGetPosRad(mor->wheels[i].steer, &steerPos[i]);
		steerPos[i] = morAngleLimit(steerPos[i]);

		rsmldt[i] = momWheel[i].radius*driveDeltaPos/servoperiod -
		    momWheel[i].arm*steerDeltaPos/servoperiod;

		vt += (i<2?1:-1) * rsmldt[i]*cos(steerPos[i]);
		k += (i<2?1:-1) *
		    (momWheel[i].arm*cos(steerPos[i]) - momWheel[i].y);
		vt += (i%2?1:-1) * rsmldt[i]*sin(steerPos[i]);
		k += (i%2?1:-1) *
		    (momWheel[i].arm*sin(steerPos[i]) + momWheel[i].x);
	}
	vt = vt/k;

	/* average linear velocity on the 4 wheels */
	vx = 0.;
	vy = 0.;
	for (i = 0; i < MOR_N_WHEELS; i++) {
		vx += rsmldt[i]*cos(steerPos[i]) -
		    (momWheel[i].arm*cos(steerPos[i]) - momWheel[i].y)*vt;
		vy += rsmldt[i]*sin(steerPos[i]) -
		    (momWheel[i].arm*sin(steerPos[i]) + momWheel[i].x)*vt;
	}
	vx /= 4.;
	vy /= 4.;

	mor->odo->vx = vx;
	mor->odo->vy = vy;
	mor->odo->w = vt;
	dt = vt * servoperiod;
	mor->odo->x += vx*cos(mor->odo->theta+dt/2)*servoperiod
	    - vy*sin(mor->odo->theta+dt/2)*servoperiod;
	mor->odo->y += vx*sin(mor->odo->theta+dt/2)*servoperiod
	    + vy*cos(mor->odo->theta+dt/2)*servoperiod;
	mor->odo->theta += dt;
}

/*
 * Compute and send speeds to the wheels
 *
 * geom->orient is the desired position of steering wheels
 * geom->speed is the desited speed of driving wheels
 */
void
morSendSpeed(MOR_STR *mor, double period)
{
	int i;
	double vsteer, vdrive;
	double steerPos;
	static double errSum[MOR_N_WHEELS] = { 0.0 };
	static double prevErr[MOR_N_WHEELS] = { 0.0 };
	double err, d;
	bool moving = false;

	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaGetPosRad(mor->wheels[i].steer, &steerPos);
		err = morAngleLimit(morAngleLimit(mor->wheels[i].geom->orient)
		    - morAngleLimit(steerPos));
		errSum[i] = 0.9*errSum[i] + 0.1*err;
		// vsteer = mor->wheels[i].geom->dorient;
		d = err - prevErr[i];
		vsteer = (0.4*err + 0.01*errSum[i] + 0.2*d) / period ;
		if (vsteer > momWheel[i].dorientmax)
			vsteer = momWheel[i].dorientmax;
		if (vsteer < -momWheel[i].dorientmax)
			vsteer = -momWheel[i].dorientmax;

		vdrive = mor->wheels[i].geom->spin
		    + momWheel[i].arm*vsteer/momWheel[i].radius
		    - couplingFactor*vsteer;

		/* steer */
#if 0
		harmonicaSetGearVelRadS(mor->wheels[i].steer, vsteer);
#else
		harmonicaSetGearPosVelRadS(mor->wheels[i].steer,
		    steerPos + err, vsteer);
#endif
		/* drive - remove speed generated by steering drive */
		harmonicaSetGearVelRadS(mor->wheels[i].drive, vdrive);
		prevErr[i] = err;
		moving = moving || ((fabs(vsteer) > MOR_SPEED_THRESHOLD)
		    || (fabs(vdrive) > MOR_SPEED_THRESHOLD));
		// printf("    -> %d %.4lf %.4lf %.4lf %.4lf\n", i, steerPos, err, errSum[i], vsteer);
	}
	mor->moving = moving;
}

bool
morSteerOk(MOR_STR *mor)
{
	int i;
	bool result =  true;
	double steerPos;

	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaGetPosRad(mor->wheels[i].steer, &steerPos);
		result &=
		    (morAngleLimit(morAngleLimit(mor->wheels[i].geom->orient)
			- morAngleLimit(steerPos)) < 0.009);
	}
	return result;
}

bool
morMoving(const MOR_STR *mor)
{
	return mor->moving;
}

/*
 * Set speeds on all motors to 0
 */
void
morStop(MOR_STR *mor)
{
	int i;

	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaSetGearVelRadS(mor->wheels[i].drive, 0);
		harmonicaSetGearVelRadS(mor->wheels[i].steer, 0);
	}
	mor->moving = false;
}

void
morMotorsOff(MOR_STR *mor)
{
	int i;

	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaStartWatchdog(mor->wheels[i].steer, false);
		harmonicaStartWatchdog(mor->wheels[i].drive, false);
		intprtSetInt(mor->wheels[i].drive, 8, 'M', 'O', 0, 0);
		intprtSetInt(mor->wheels[i].steer, 8, 'M', 'O', 0, 0);
	}
	mor->moving = false;
}

void
morMotorsOn(MOR_STR *mor)
{
	int i;

	for (i = 0; i < MOR_N_WHEELS; i++) {
		intprtSetInt(mor->wheels[i].drive, 8, 'M', 'O', 0, 1);
		intprtSetInt(mor->wheels[i].steer, 8, 'M', 'O', 0, 1);
		harmonicaStartWatchdog(mor->wheels[i].steer, true);
		harmonicaStartWatchdog(mor->wheels[i].drive, true);
	}
}


int
morStartReadTask(MOR_STR *mor)
{

	if (pthread_create(&mor->readTask, NULL, morReadTask, mor) != 0)
		return -1;
	return 0;
}


static void *
morReadTask(void *p)
{
	MOR_STR *mor = (MOR_STR *)p;

	do {
		while (morEvalCanBuffer(mor, 10) == 0)
			;
	} while (!mor->dying);

	return NULL;
}

static char *anim[] = {
	"---(o           )---",
	"---( o          )---",
	"---(  o         )---",
	"---(   o        )---",
	"---(    o       )---",
	"---(     o      )---",
	"---(      o     )---",
	"---(       o    )---",
	"---(        o   )---",
	"---(         o  )---",
	"---(          o )---",
	"---(           o)---",
	"---(          o )---",
	"---(         o  )---",
	"---(        o   )---",
	"---(       o    )---",
	"---(      o     )---",
	"---(     o      )---",
	"---(    o       )---",
	"---(   o        )---",
	"---(  o         )---",
	"---( o          )---",
};

static void *
morRelayboardTask(void *p)
{
	struct tm tm;
	MOR_STR *mor = (MOR_STR *)p;
	RELAYBOARD_STR *rb = mor->relayboard;
	struct mortimer *timer;
	char buffer[128];
	time_t now;
	int i = 0;

	timer = mortimerAlloc();
	if (mortimerStart(timer, 50, 50) == -1) {
		warn("%s: mortimerStart", __func__);

		return NULL;
	}
	while (!mor->dying) {
		mortimerPause(timer);
		time(&now);
		localtime_r(&now, &tm);
		strftime(buffer, sizeof(buffer),
		    "%F %T "
		    "* CNRS/LAAS  Ready *",
		    &tm);
		sprintf(buffer+40, "%s", anim[i]);
		i = (i+1) % 22;
		relayboardUpdate(rb, buffer);
	}
	return NULL;
}


int
morStartRelayboardTask(MOR_STR *mor)
{
	if (pthread_create(&mor->relayboardTask, NULL,
		morRelayboardTask, mor) != 0)
		return -1;
	return 0;
}
