/*
 * Copyright (c) 2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** Driver for the Neobotix MOR relayboard, serial connection
 **/
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include <linux/serial.h>

#include "mor/relayboard.h"

#define RELAYBOARD_BAUDRATE 420000
#define RELAYBOARD_RQST_LEN 79
#define RELAYBOARD_DATA_LEN 110

static void relayboardDecode(RELAYBOARD_STR *);
static void relayboardSendRequest(RELAYBOARD_STR *, const char *);

/**
 * Init communication with the relayboard
 *
 * input: devname - the serial device connecting to the relaboard
 */
RELAYBOARD_STR *
relayboardInit(const char *devname)
{
	RELAYBOARD_STR *rb;
	struct termios tio;
	struct serial_struct ss;
	int fd;

	rb = (RELAYBOARD_STR *)calloc(1, sizeof(RELAYBOARD_STR));
	if (rb == NULL)
		return NULL;

	fd = open(devname, O_RDWR | O_NOCTTY);
	if (fd < 0) {
		free(rb);
		return NULL;
	}

	/* set raw mode */
	if (tcgetattr(fd, &tio) == -1) {
		relayboardEnd(rb);
		return NULL;
	}
	cfmakeraw(&tio);
	/* blocking mode */
	tio.c_cc[VTIME] = 1;
	tio.c_cc[VMIN] = 0;
	/* no flow control */
	tio.c_cflag &= ~CRTSCTS;
	tio.c_iflag &= ~(IXON | IXOFF | IXANY);
	cfsetspeed(&tio, B38400);
	if (tcsetattr(fd, TCSANOW, &tio) == -1) {
		relayboardEnd(rb);
		return NULL;
	}
	/* specific speed */
	if (ioctl(fd, TIOCGSERIAL, &ss) == -1) {
		relayboardEnd(rb);
		return NULL;
	}
	ss.flags |= ASYNC_SPD_CUST;
	ss.custom_divisor = ss.baud_base / RELAYBOARD_BAUDRATE;
	if (ioctl(fd, TIOCSSERIAL, &ss) == -1) {
		relayboardEnd(rb);
		return NULL;
	}
	rb->fd = fd;
	return rb;
}

/**
 * End commuication with relayboard and release data
 */
void
relayboardEnd(RELAYBOARD_STR *rb)
{
	if (rb == NULL)
		return;
	close(rb->fd);
	free(rb);
}

/**
 * Print current relayboard data to stdout
 */
void
relayboardPrint(RELAYBOARD_STR *rb)
{
	printf("board status: %04x\n", rb->state);
	printf("charging current: %d\n", rb->charge_c);
	printf("voltage: %.3f V\n", rb->voltage/1000.0);
	printf("Temperature: %d deg C\n", rb->temperature);
	printf("keypad: %04x\n", rb->keypad);
}

/**
 * Update data from relay board
 *
 * input : message: new string to display on the robot's LCD screen
 *                  NULL -> don't update it
 */
int
relayboardUpdate(RELAYBOARD_STR *rb, const char *message)
{
	ssize_t nread, n = 0;
	int nready;

	relayboardSendRequest(rb, message);
	do {
		n = 0;
		while (n < RELAYBOARD_DATA_LEN) {
			errno = 0;
			nread = read(rb->fd, rb->buf + n,
			    RELAYBOARD_DATA_LEN - n);
			if (nread < 0 && errno == EAGAIN)
				continue;
			if (nread <= 0) {
				// warn("read error %d", nread);
				break;
			}
			n += nread;
		}

		if (n != RELAYBOARD_DATA_LEN) {
			// warn("read: %zd bytes read", n);
			return -1;
		}
		if (ioctl(rb->fd, FIONREAD, &nready) == -1)
			warn("FIONREAD");
		// printf("nready %d\n", nready);
	} while (nready > 0);
	relayboardDecode(rb);
	return 0;
}

/*----------------------------------------------------------------------*/

/**
 * Decode a data packet from the relayboard
 */
static void
relayboardDecode(RELAYBOARD_STR *rb)
{
	unsigned char *buf = rb->buf;

	if (buf[0] != 0x02 || buf[1] != 0x80 ||
	    buf[2] != 0xd6 || buf[3] != 0x02) {
		warnx("invalid header %02x %02x %02x %02x\n",
		    buf[0], buf[1], buf[2], buf[3]);
	}
	rb->state = buf[4] | (buf[5] << 8);
	rb->charge_c = buf[6] + 256*(signed char)buf[7];
	rb->voltage = buf[8] + 256*(signed char)buf[9];
	rb->keypad = buf[10] | (buf[11] << 8);
	rb->temperature = buf[20] + 256*(signed char)buf[21];
}

/**
 * Send an uppdate request to the relayboard
 */
static void
relayboardSendRequest(RELAYBOARD_STR *rb, const char *message)
{
	unsigned char *buf = rb->buf;
	int i = 0, sum = 0;

	memset(rb->buf, 0, sizeof(rb->buf));
	buf[i++] = 1;		/* SET_GET_DATA */
	buf[i++] = 0;		/* config 2 bytes */
	buf[i++] = 0;
	buf[i++] = 0;		/* commands 2 bytes */
	buf[i++] = 0;
	if (message != NULL) {
		snprintf((char *)buf+17, 20, "%20s", message);
	}
	/* checksum */
	for (i = 0; i < RELAYBOARD_RQST_LEN - 2; i++)
		sum += buf[i];
	buf[77] = sum >> 8;
	buf[78] = sum & 0xff;
	tcflush(rb->fd, TCOFLUSH);
	if (write(rb->fd, rb->buf, RELAYBOARD_RQST_LEN) != RELAYBOARD_RQST_LEN)
		warn("%s", __func__);
}
