/*
 * Copyright (c) 2009,2010,2011,2018 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** Program to help calibrating the mechanical offsets of the steering
 ** wheels of the MOR platform
 **
 ** principle:
 **  1. set encOffsetIncr = 0 for all steering wheels in ../src/mom.h
 **  2. run the program and let the robot do the homing of the 
 **     steering wheels
 **  3. once the wheel are at the homing position, use a ruler placed 
 **     against the wheels to align them two by two.
 **  4. record the offsets that are displayed
 **     this may be repeated a few times to get a better estimation 
 **     by taking  the mean values for each wheel
 **  5. set encOffsetIncr to the corresponding value for each wheel. 
 **     /!\ for wheels 3 and 4, you need to add PI (25941 encoder steps) 
 **     /!\ take the sign of the rotation for each steering weel into 
 **         account. 
 **  6. rebuild morCalib with the new values, run it and re-do the 
 **     alignment.  Normally the displayed offsets  should be near 0.
 **     If not, check for sign errors.
 **  7. the '#if 0' below can be changed to '#if 1' to let the wheels
 **     move to their rest positions. The wheels should be aligned.
 **/
#include <pthread.h>

#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <poll.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mor/socketcan.h"
#include "mor/harmonica.h"
#include "mor/relayboard.h"
#include "mor/morLib.h"


double vx, vy, w;		/* linear & angular speed references */
double s;			/* spine speed */
int quit;			/* quit flag */

#define RAD_TO_DEG(x) ((x)*180/M_PI)

/*----------------------------------------------------------------------*/
/**
 ** Main loop of  control
 **/
static void
control(MOR_STR *mor)
{
	CAN_HARMONICA_STR *spine = mor->spine;
	int i = 0;

	morMotorsOff(mor);
	while (!quit) {

		// morMove(mor, vx, vy, w, 0.05);
		// morSendSpeed(mor, 0.05);
		harmonicaSetGearVelRadS(spine, 0);
		harmonicaRequestDigIn(spine);
		usleep(500000);
		while (morEvalCanBuffer(mor, 0) == 0)
			;
		for (i = 0; i < MOR_N_WHEELS; i++) {
			CAN_HARMONICA_STR *h = mor->wheels[i].steer;
			double angle;

			harmonicaGetPosRad(h, &angle);
			printf("wheel %d: %.2lf deg ", i+1, RAD_TO_DEG(angle));
			printf("%d steps\n",
			    driveParamPosGearRadToPosMotIncr(&h->driveParam,
				angle));
		}
		printf("\n");

	} /* while */
	relayboardUpdate(mor->relayboard, "                    ");
	if (mor->relayboard->state == 0x1) {
		printf("E-Stop\n");
		quit++;
	}
	morStop(mor);
} /* joymove */


/*----------------------------------------------------------------------*/
int
main(int argc, char *argv[])
{
	MOR_STR *mor;

	vx = 0;
	vy = 0;
	w = 0;
	quit = 0;

	/* Connect to MOR devices */
	mor = morInit("can0", "/dev/ttyUSB0", true, true);
	if (mor == NULL)
		exit(2);

	printf("morInit done\n");
	if (morInitPlatform(mor) == -1) {
		printf("initPlatform failed\n");
		morEnd(mor);
		exit(2);
	}
	printf("start homing\n");
	morHoming(mor);
	for (int i = 0; i < MOR_N_WHEELS; i++) {
		printf("%.2lf ", mor->wheels[i].geom->orient);
	}
	printf("\n");
#if 0
	printf("wait\n");
	// morMove(mor, 0, 0, 0, 0.05);
	for (int i = 0; i < MOR_N_WHEELS; i++)
		mor->wheels[i].geom->orient = i < 2 ?  M_PI : 0;

	for (int t = 0; t < 100; t++) {
		harmonicaSetGearPosVelRadS(mor->spine, 0.00, 0.05);
		morSendSpeed(mor, 0.05);
		usleep(50000);
		while (morEvalCanBuffer(mor, 0) == 0)
			;
	}
#endif
	/* launch main loop */
	control(mor);

	/* Free resources */
	morEnd(mor);
	return 0;
}
