/*
 * Copyright (c) 2009,2010,2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>

#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <poll.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <linux/joystick.h>

#include "mor/socketcan.h"
#include "mor/harmonica.h"
#include "mor/relayboard.h"
#include "mor/morLib.h"

/* For joystick driver */
#define NAME_LENGTH 128

unsigned char axes = 2;		/* number of joystick axes */
unsigned char buttons = 2;	/* number of joystick buttons */
int *axis;			/* array of joystick axis values */
int *button;			/* array of joystick button state */
double vx, vy, w;		/* linear & angular speed references */
double s;			/* spine speed */
int quit;			/* quit flag */

int version = 0x000800;
char name[NAME_LENGTH] = "Unknown";

pthread_mutex_t joymutex = PTHREAD_MUTEX_INITIALIZER;


/* Speed gains for joystick control */
#define VGAIN	-0.3/(32768.0) /* linear  */
#define WGAIN   -0.5/(32768.0) /* angular */
#define SGAIN   -0.07/(32768.0) /* spine */

/*----------------------------------------------------------------------*/
/**
 ** Init joystick
 **/
static int
joyinit(char *dev)
{
	int joy;

	joy = open(dev, O_RDONLY);
	if (joy < 0) {
		perror(dev);
		return -1;
	}
	ioctl(joy, JSIOCGVERSION, &version);
	ioctl(joy, JSIOCGAXES, &axes);
	ioctl(joy, JSIOCGBUTTONS, &buttons);
	ioctl(joy, JSIOCGNAME(NAME_LENGTH), name);

	printf("Joystick (%s) has %d axes and %d buttons.\n"
	       "\tDriver version is %d.%d.%d.\n",
	       name, axes, buttons, version >> 16,
	       (version >> 8) & 0xff, version & 0xff);
	/* flush */
	return joy;
}
/*----------------------------------------------------------------------*/

/**
 ** task reading data from the joystick and producing references;
 **/
void *
joyRead(void *ptr)
{
	struct js_event js;
	struct pollfd pfd;
	int fd = *(int *)ptr;
	int ret, quitting = 0;

	while (!quit) {
		pfd.fd = fd;
		pfd.events = POLLIN;
		pfd.revents = 0;
		ret = poll(&pfd, 1, 100);

		if (ret < 0) {
			if (errno != EINTR)
				fprintf(stderr, "poll error: %s\n",
					strerror(errno));
			continue;
		}
		if (ret == 0)
			/* nothing to read from joystick */
			continue;
		if ((pfd.revents & POLLIN) != POLLIN)
			continue;
		/* something to read from joystick */
		if (read(fd, &js, sizeof(struct js_event))
		    != sizeof(struct js_event)) {
			perror("\njstest: error reading");
			break;
		}
		switch(js.type & ~JS_EVENT_INIT) {
		case JS_EVENT_BUTTON:
			button[js.number] = js.value;
			//printf("button %d %d\n", js.number, js.value);
			break;
		case JS_EVENT_AXIS:
			axis[js.number] = js.value;
			//printf("axis %d %d\n", js.number, js.value);
			break;
		}
		pthread_mutex_lock(&joymutex);
		/* Set refs to zero */
		vx = 0;
		vy = 0;
		w = 0;
		s = 0;
		/* Test for motion button */
		if (button[0]) {
			/* compute speed refs */
			vx = axis[1]*VGAIN;
			vy = 0.0;
			w = axis[0]*WGAIN;
		} else if (button[1]) {
			vx = axis[1]*VGAIN;
			vy = axis[0]*VGAIN;
			w = axis[2]*WGAIN;
		} else if (button[2]) {
			s = axis[1]*SGAIN;
		}
		if (button[3])
			quitting = 1;
		if (!button[3] && quitting)
			quit++;
		pthread_mutex_unlock(&joymutex);
	}
	return NULL;
}

bool prevSpeedIsNull = true;

static void
checkMotion(MOR_STR *mor)
{
	if (button[0] == 0 && button[1] == 0 && morSteerOk(mor)) {
		if (!prevSpeedIsNull) {
			printf(" -- motors OFF\n");
			morMotorsOff(mor);
			prevSpeedIsNull = true;
		}
	} else {
		if (prevSpeedIsNull) {
			printf(" -- motors ON\n");
			morMotorsOn(mor);
			prevSpeedIsNull = false;
		}
	}
}

/*----------------------------------------------------------------------*/
/**
 ** Main loop of joystick control mode
 **/
static void
joymove(MOR_STR *mor)
{
	CAN_HARMONICA_STR *spine = mor->spine;
	int i = 0;

	while (!quit) {
		pthread_mutex_lock(&joymutex);
		checkMotion(mor);
		morMove(mor, vx, vy, w, 0.05);
		morSendSpeed(mor, 0.05);
#ifndef WITHOUT_SPINE
		if ((spine->digIn & HARMONICA_INPUT_FORWARD_LIMIT) != 0 
		    && s > 0) {
			printf("upper spine limit reached\n");
			s = 0.;
		}
		if ((spine->digIn & HARMONICA_INPUT_REVERSE_LIMIT) !=0
		    && s < 0) {
			printf("lower spine limit reached\n");
			s = 0.;
		}
		harmonicaSetGearVelRadS(spine, s);
		harmonicaRequestDigIn(spine);
#endif
		// retval = morRotate(mor, r);
		pthread_mutex_unlock(&joymutex);
		if (++i == 25) {
			relayboardUpdate(mor->relayboard,
			    "**  JOYSTICK  ON  **");
			if (mor->relayboard->state == 0x1) {
				printf("E-Stop\n");
				quit++;
			}
			i = 0;
		}
		usleep(50000);
		while (morEvalCanBuffer(mor, 0) == 0)
			;
		//if (mor->motorFailure)
		//	quit++;
		morOdo(mor, 0.05);
		printf("-- %.4lf %.4lf %.4lf\n",
		    mor->odo->x, mor->odo->y, mor->odo->theta);
		
	} /* while */
	relayboardUpdate(mor->relayboard, "                    ");
	if (mor->relayboard->state == 0x1) {
		printf("E-Stop\n");
	}
	morStop(mor);
} /* joymove */


/*----------------------------------------------------------------------*/
int
main(int argc, char *argv[])
{
	MOR_STR *mor;
	pthread_t t;
	int j;

	vx = 0;
	vy = 0;
	w = 0;
	quit = 0;


	/* Connect to joystick */
	j = joyinit("/dev/input/js0");
	if (j < 0) {
		fprintf(stderr, "Error initializing joystick\n");
		exit(2);
	}
	printf("joyinit done\n");
	axis = calloc(axes, sizeof(int));
	button = calloc(buttons, sizeof(int));

	pthread_create(&t, NULL, joyRead, &j);
	pthread_detach(t);

	sleep(1);

	/* Connect to MOR devices */
	mor = morInit("can0", "/dev/ttyUSB0", true, true);
	if (mor == NULL)
		exit(2);

	printf("morInit done\n");
	if (morInitPlatform(mor) == -1) {
		printf("initPlatform failed\n");
		morEnd(mor);
		exit(2);
	}
	printf("start homing\n");
	morHoming(mor);
#ifndef WITHOUT_SPINE
	printf("move spine\n");
	do {
		harmonicaSetGearPosVelRadS(mor->spine, 0.02, 0.05);
		morSendSpeed(mor, 0.05);
		usleep(50000);
		while (morEvalCanBuffer(mor, 0) == 0)
			;
		printf("-- %.2lf\n", mor->spine->posGearMeasRad);
	} while (fabs(mor->spine->posGearMeasRad - 0.02) > 0.01);
#endif
	printf("Use button 1 and left pad for non-holonomic motion\n");
	printf("    button 2 and left pad (x/y) + right pad (theta)\n");
	printf("           for omni-directional motion\n");
	printf("    button 3 and left pad for spine motion\n");
	printf("    button 4 to quit\n");

	/* launch main loop */
	joymove(mor);

	/* Free resources */
	free(button);
	free(axis);
	close (j);
	morEnd(mor);
	return 0;
}
