/*
 * Copyright (c) 2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <err.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <mor/socketcan.h>
#include <mor/harmonica.h>


/* Spine motor */
const struct CAN_OPEN_PARAM_STR spineAddr = {
	.txPDO1 = 0x189,
	.txPDO2 = 0x289,
	.rxPDO2 = 0x309,
	.txSDO = 0x589,
	.rxSDO = 0x609
};

/*
 * For this motor, the position and speed are in m and m/s respectively
 */
const struct CAN_DRIVE_PARAMS spineParams = {
	.encIncrPerRevMot = 4096,
	.gearRatio = 1870,	/* conversion factor to get meters
				   from encoders */
	.beltRatio = 1,
	.sign = 1.0,
	.velMaxEncIncrS = 300000,
	.accIncrS = 300000,
	.decIncrS = 300000,
	.encOffsetIncr = 10000,
	.isSteering = true,
	.currToTorque = 0.10065,
	.currMax = 9.62,
	.homingDigIn = 7,
	.homingSpeed = -20000,	/* move downwards for calibration */
	.homingDelay = 40,	/* theorical 25s at maximum course */
};

int
main(int argc, char *argv[])
{
	CAN_MSG m;
	int fd;
	CAN_HARMONICA_STR *h;
	char *name;

	if (argc == 2)
		name = argv[1];
	else
		name = "can0";

	fd = socketcanInit(name);
	if (fd == -1)
		err(2, "%s", name);
	
	h = harmonicaInit(fd, 0, MOTIONTYPE_POSCTRL,
	    &spineAddr, &spineParams);

	if (h == NULL) {
		socketcanEnd(fd);
		err(2, "harmonicaInit");
	}
	/* init the watchdog, but disable it */
	harmonicaStartWatchdog(h, true);
	usleep(10000);
	harmonicaStartWatchdog(h, false);
	printf("watchdog initialized\n");


	/* Init motor */
	if (harmonicaInitCtrl(h) == -1) {
		return -1;
	}
	printf("harmonica control started\n");

	/* Start motor */
	if (harmonicaStart(h) == -1) {
		return -1;
	}
	/* To be sure */
	harmonicaSetGearVelRadS(h, 0.0);
	printf("motor started\n");

	harmonicaInitHoming(h);
	if (harmonicaExecHoming(h) == -1) {
		socketcanEnd(fd);
		exit(-2);
	}
	while (socketcanReceiveMsg(h->dev, &m) == 0);

	harmonicaStartWatchdog(h, true);
	while ((h->digIn & HARMONICA_INPUT_FORWARD_LIMIT) == 0) {
		harmonicaSetGearPosVelRadS(h, 0.4, 0.03);
		harmonicaRequestDigIn(h);
		harmonicaRequestCurrent(h);
		while (socketcanReceiveMsg(h->dev, &m) == 0) {
			harmonicaDecode(h, &m);
		}
		//harmonicaShow(h);
		usleep(100000);
	}
	printf("Reached top\n");
	harmonicaStartWatchdog(h, false);
	harmonicaSetGearPosVelRadS(h, 0.16, -0.02);
#if 0
	while ((h->digIn & HARMONICA_INPUT_REVERSE_LIMIT) == 0) {
		harmonicaRequestDigIn(h);
		harmonicaRequestCurrent(h);
		while (socketcanReceiveMsg(h->dev, &m) == 0) {
			harmonicaDecode(h, &m);
		}
		//harmonicaShow(h);
		usleep(100000);
	}
	harmonicaSetGearVelRadS(h, 0.0);
#endif

	socketcanEnd(fd);
	exit(0);
}
